"""Tests for the `physbiblio.webimport` package.

This file is part of the physbiblio package.
"""
__author__ = 'Stefano Gariazzo'
__email__ = "stefano.gariazzo@gmail.com"

import physbiblio
__version__ = physbiblio.__version__
__version_date__ = physbiblio.__version_date__

__all__ = ["test_webimport"]
