"""Module for retrieving bibtex information
from a list of given web resources
(see `__all__`).

This module is part of the physbiblio package.
"""
__author__ = 'Stefano Gariazzo'
__email__ = "stefano.gariazzo@gmail.com"

import physbiblio
__version__ = physbiblio.__version__
__version_date__ = physbiblio.__version_date__

__all__ = [
	"arxiv",
	"doi",
	"inspire",
	"inspireoai",
	"isbn",
	"webInterf",
	"tests",
	]
